@@ -40,8 +40,10 @@
 #include "common/help.h"
 #include "common/units.h"
 #include "common/parse-utils.h"
+#include "common/string-utils.h"
 #include "common/format-output.h"
 #include "common/messages.h"
+#include "common/tree-search.h"
 #include "cmds/commands.h"
 #include "cmds/qgroup.h"
 
@@ -1254,13 +1256,12 @@ static bool key_in_range(const struct btrfs_key *key,
 	return true;
 }
 
-static int __qgroups_search(int fd, struct btrfs_ioctl_search_args *args,
+static int __qgroups_search(int fd, struct btrfs_tree_search_args *args,
 			    struct qgroup_lookup *qgroup_lookup)
 {
 	int ret;
-	struct btrfs_ioctl_search_key *sk = &args->key;
-	struct btrfs_ioctl_search_key filter_key = args->key;
-	struct btrfs_ioctl_search_header *sh;
+	struct btrfs_ioctl_search_key *sk;
+	struct btrfs_ioctl_search_key filter_key;
 	unsigned long off = 0;
 	unsigned int i;
 	struct btrfs_qgroup_status_item *si;
@@ -1270,10 +1271,13 @@ static int __qgroups_search(int fd, struct btrfs_ioctl_search_args *args,
 	u64 qgroupid;
 	u64 child, parent;
 
+	sk = btrfs_tree_search_sk(args);
+	filter_key = *sk;
+
 	qgroup_lookup_init(qgroup_lookup);
 
 	while (1) {
-		ret = ioctl(fd, BTRFS_IOC_TREE_SEARCH, args);
+		ret = btrfs_tree_search_ioctl(fd, args);
 		if (ret < 0) {
 			if (errno == ENOENT)
 				ret = -ENOTTY;
@@ -1293,38 +1297,35 @@ static int __qgroups_search(int fd, struct btrfs_ioctl_search_args *args,
 		 */
 		for (i = 0; i < sk->nr_items; i++) {
 			struct btrfs_key key;
+			struct btrfs_ioctl_search_header sh;
 
-			sh = (struct btrfs_ioctl_search_header *)(args->buf +
-								  off);
-			off += sizeof(*sh);
+			memcpy(&sh, btrfs_tree_search_data(args, off), sizeof(sh));
+			off += sizeof(sh);
 
-			key.objectid = btrfs_search_header_objectid(sh);
-			key.type = btrfs_search_header_type(sh);
-			key.offset = btrfs_search_header_offset(sh);
+			key.objectid = sh.objectid;
+			key.type = sh.type;
+			key.offset = sh.offset;
 
 			if (!key_in_range(&key, &filter_key))
 				goto next;
 
 			switch (key.type) {
 			case BTRFS_QGROUP_STATUS_KEY:
-				si = (struct btrfs_qgroup_status_item *)
-				     (args->buf + off);
+				si = btrfs_tree_search_data(args, off);
 				flags = btrfs_stack_qgroup_status_flags(si);
 
 				print_status_flag_warning(flags);
 				break;
 			case BTRFS_QGROUP_INFO_KEY:
 				qgroupid = key.offset;
-				info = (struct btrfs_qgroup_info_item *)
-				       (args->buf + off);
+				info = btrfs_tree_search_data(args, off);
 
 				ret = update_qgroup_info(fd, qgroup_lookup,
 							 qgroupid, info);
 				break;
 			case BTRFS_QGROUP_LIMIT_KEY:
 				qgroupid = key.offset;
-				limit = (struct btrfs_qgroup_limit_item *)
-					(args->buf + off);
+				limit = btrfs_tree_search_data(args, off);
 
 				ret = update_qgroup_limit(fd, qgroup_lookup,
 							  qgroupid, limit);
@@ -1347,7 +1348,7 @@ static int __qgroups_search(int fd, struct btrfs_ioctl_search_args *args,
 				return ret;
 
 next:
-			off += btrfs_search_header_len(sh);
+			off += sh.len;
 
 			/*
 			 * record the mins in sk so we can make sure the
@@ -1373,19 +1374,20 @@ next:
 
 static int qgroups_search_all(int fd, struct qgroup_lookup *qgroup_lookup)
 {
-	struct btrfs_ioctl_search_args args = {
-		.key = {
-			.tree_id = BTRFS_QUOTA_TREE_OBJECTID,
-			.max_type = BTRFS_QGROUP_RELATION_KEY,
-			.min_type = BTRFS_QGROUP_STATUS_KEY,
-			.max_objectid = (u64)-1,
-			.max_offset = (u64)-1,
-			.max_transid = (u64)-1,
-			.nr_items = 4096,
-		},
-	};
+	struct btrfs_tree_search_args args;
+	struct btrfs_ioctl_search_key *sk;
 	int ret;
 
+	memset(&args, 0, sizeof(args));
+	sk = btrfs_tree_search_sk(&args);
+	sk->tree_id = BTRFS_QUOTA_TREE_OBJECTID;
+	sk->max_type = BTRFS_QGROUP_RELATION_KEY;
+	sk->min_type = BTRFS_QGROUP_STATUS_KEY;
+	sk->max_objectid = (u64)-1;
+	sk->max_offset = (u64)-1;
+	sk->max_transid = (u64)-1;
+	sk->nr_items = 4096;
+
 	ret = __qgroups_search(fd, &args, qgroup_lookup);
 	if (ret == -ENOTTY) {
 		error("can't list qgroups: quotas not enabled");
@@ -1398,23 +1400,24 @@ static int qgroups_search_all(int fd, struct qgroup_lookup *qgroup_lookup)
 
 int btrfs_qgroup_query(int fd, u64 qgroupid, struct btrfs_qgroup_stats *stats)
 {
-	struct btrfs_ioctl_search_args args = {
-		.key = {
-			.tree_id = BTRFS_QUOTA_TREE_OBJECTID,
-			.min_type = BTRFS_QGROUP_INFO_KEY,
-			.max_type = BTRFS_QGROUP_LIMIT_KEY,
-			.max_objectid = 0,
-			.min_offset = qgroupid,
-			.max_offset = qgroupid,
-			.max_transid = (u64)-1,
-			.nr_items = 4096,
-		},
-	};
+	struct btrfs_tree_search_args args;
+	struct btrfs_ioctl_search_key *sk;
 	struct qgroup_lookup qgroup_lookup;
 	struct btrfs_qgroup *qgroup;
 	struct rb_node *n;
 	int ret;
 
+	memset(&args, 0, sizeof(args));
+	sk = btrfs_tree_search_sk(&args);
+	sk->tree_id = BTRFS_QUOTA_TREE_OBJECTID;
+	sk->min_type = BTRFS_QGROUP_INFO_KEY;
+	sk->max_type = BTRFS_QGROUP_LIMIT_KEY;
+	sk->max_objectid = 0;
+	sk->min_offset = qgroupid;
+	sk->max_offset = qgroupid;
+	sk->max_transid = (u64)-1;
+	sk->nr_items = 4096;
+
 	ret = __qgroups_search(fd, &args, &qgroup_lookup);
 	if (ret < 0)
 		return ret;
@@ -1641,7 +1644,7 @@ static int qgroup_inherit_realloc(struct btrfs_qgroup_inherit **inherit, int n,
 			 (*inherit)->num_excl_copies;
 	}
 
-	out = calloc(sizeof(*out) + sizeof(out->qgroups[0]) * (nitems + n), 1);
+	out = calloc(1, sizeof(*out) + sizeof(out->qgroups[0]) * (nitems + n));
 	if (out == NULL) {
 		error_msg(ERROR_MSG_MEMORY, NULL);
 		return -ENOMEM;
@@ -1686,47 +1689,6 @@ int btrfs_qgroup_inherit_add_group(struct btrfs_qgroup_inherit **inherit, char *
 	return 0;
 }
 
-int btrfs_qgroup_inherit_add_copy(struct btrfs_qgroup_inherit **inherit, char *arg,
-			    int type)
-{
-	int ret;
-	u64 qgroup_src;
-	u64 qgroup_dst;
-	char *p;
-	int pos = 0;
-
-	p = strchr(arg, ':');
-	if (!p) {
-bad:
-		error("invalid copy specification, missing separator :");
-		return -EINVAL;
-	}
-	*p = 0;
-	qgroup_src = parse_qgroupid_or_path(arg);
-	qgroup_dst = parse_qgroupid_or_path(p + 1);
-	*p = ':';
-
-	if (!qgroup_src || !qgroup_dst)
-		goto bad;
-
-	if (*inherit)
-		pos = (*inherit)->num_qgroups +
-		      (*inherit)->num_ref_copies * 2 * type;
-
-	ret = qgroup_inherit_realloc(inherit, 2, pos);
-	if (ret)
-		return ret;
-
-	(*inherit)->qgroups[pos++] = qgroup_src;
-	(*inherit)->qgroups[pos++] = qgroup_dst;
-
-	if (!type)
-		++(*inherit)->num_ref_copies;
-	else
-		++(*inherit)->num_excl_copies;
-
-	return 0;
-}
 static const char * const qgroup_cmd_group_usage[] = {
 	"btrfs qgroup <command> [options] <path>",
 	NULL
@@ -1740,7 +1702,6 @@ static int _cmd_qgroup_assign(const struct cmd_struct *cmd, int assign,
 	bool rescan = true;
 	char *path;
 	struct btrfs_ioctl_qgroup_assign_args args;
-	DIR *dirstream = NULL;
 
 	optind = 0;
 	while (1) {
@@ -1785,7 +1746,7 @@ static int _cmd_qgroup_assign(const struct cmd_struct *cmd, int assign,
 		error("bad relation requested: %s", path);
 		return 1;
 	}
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0)
 		return 1;
 
@@ -1794,7 +1755,7 @@ static int _cmd_qgroup_assign(const struct cmd_struct *cmd, int assign,
 		error("unable to assign quota group: %s",
 				errno == ENOTCONN ? "quota not enabled"
 						: strerror(errno));
-		close_file_or_dir(fd, dirstream);
+		close(fd);
 		return 1;
 	}
 
@@ -1820,7 +1781,7 @@ static int _cmd_qgroup_assign(const struct cmd_struct *cmd, int assign,
 			ret = 0;
 		}
 	}
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	return ret;
 }
 
@@ -1830,7 +1791,6 @@ static int _cmd_qgroup_create(int create, int argc, char **argv)
 	int fd;
 	char *path;
 	struct btrfs_ioctl_qgroup_create_args args;
-	DIR *dirstream = NULL;
 
 	if (check_argc_exact(argc - optind, 2))
 		return 1;
@@ -1844,12 +1804,12 @@ static int _cmd_qgroup_create(int create, int argc, char **argv)
 	}
 	path = argv[optind + 1];
 
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0)
 		return 1;
 
 	ret = ioctl(fd, BTRFS_IOC_QGROUP_CREATE, &args);
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	if (ret < 0) {
 		error("unable to %s quota group: %s",
 			create ? "create":"destroy",
@@ -1957,7 +1917,6 @@ static int cmd_qgroup_show(const struct cmd_struct *cmd, int argc, char **argv)
 	char *path;
 	int ret = 0;
 	int fd;
-	DIR *dirstream = NULL;
 	u64 qgroupid;
 	int filter_flag = 0;
 	unsigned unit_mode;
@@ -2030,7 +1989,7 @@ static int cmd_qgroup_show(const struct cmd_struct *cmd, int argc, char **argv)
 		return 1;
 
 	path = argv[optind];
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0) {
 		free(filter_set);
 		free(comparer_set);
@@ -2048,7 +2007,7 @@ static int cmd_qgroup_show(const struct cmd_struct *cmd, int argc, char **argv)
 		if (ret < 0) {
 			errno = -ret;
 			error("cannot resolve rootid for %s: %m", path);
-			close_file_or_dir(fd, dirstream);
+			close(fd);
 			goto out;
 		}
 		if (filter_flag & 0x1)
@@ -2061,7 +2020,7 @@ static int cmd_qgroup_show(const struct cmd_struct *cmd, int argc, char **argv)
 					qgroupid);
 	}
 	ret = show_qgroups(fd, filter_set, comparer_set);
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	free(filter_set);
 	free(comparer_set);
 
@@ -2088,7 +2047,6 @@ static int cmd_qgroup_limit(const struct cmd_struct *cmd, int argc, char **argv)
 	unsigned long long size;
 	bool compressed = false;
 	bool exclusive = false;
-	DIR *dirstream = NULL;
 	enum btrfs_util_error err;
 
 	optind = 0;
@@ -2114,7 +2072,7 @@ static int cmd_qgroup_limit(const struct cmd_struct *cmd, int argc, char **argv)
 	if (!strcasecmp(argv[optind], "none"))
 		size = -1ULL;
 	else
-		size = parse_size_from_string(argv[optind]);
+		size = arg_strtou64_with_suffix(argv[optind]);
 
 	memset(&args, 0, sizeof(args));
 	if (compressed)
@@ -2146,12 +2104,12 @@ static int cmd_qgroup_limit(const struct cmd_struct *cmd, int argc, char **argv)
 	} else
 		usage(cmd, 1);
 
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0)
 		return 1;
 
 	ret = ioctl(fd, BTRFS_IOC_QGROUP_LIMIT, &args);
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	if (ret < 0) {
 		error("unable to limit requested quota group: %s",
 				errno == ENOTCONN ? "quota not enabled"
@@ -2179,7 +2137,6 @@ static int cmd_qgroup_clear_stale(const struct cmd_struct *cmd, int argc, char *
 	int ret = 0;
 	int fd;
 	char *path = NULL;
-	DIR *dirstream = NULL;
 	struct qgroup_lookup qgroup_lookup;
 	struct rb_node *node;
 	struct btrfs_qgroup *entry;
@@ -2189,7 +2146,7 @@ static int cmd_qgroup_clear_stale(const struct cmd_struct *cmd, int argc, char *
 
 	path = argv[optind];
 
-	fd = btrfs_open_dir(path, &dirstream, 1);
+	fd = btrfs_open_dir(path);
 	if (fd < 0)
 		return 1;
 
@@ -2225,7 +2182,7 @@ static int cmd_qgroup_clear_stale(const struct cmd_struct *cmd, int argc, char *
 	}
 
 out:
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	return !!ret;
 }
 static DEFINE_SIMPLE_COMMAND(qgroup_clear_stale, "clear-stale");
